# -*- coding: utf-8 -*-
""" Non-specification rules.

This module contains information about where to find the RFC specifications
as well as what encoding rule to utilize when decoding/encoding bytes/strings.
"""

# IETF Website
IETF = "https://tools.ietf.org/html/"

# Encoding Rule
ENCODE = "ascii"
